# MIT License

# Copyright (c) 2018 Unplugged Coder

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import logging
from concurrent import futures


class AsyncLogDispatcher(logging.Handler):
    ''' A dispatcher class which process logging in an asynchronous way.

    It doesn't do any log procedure actually. Instead, it just provide the
    asynchonrous way to do the logging. That's why it named as Dispatcher.
    It's useful when logging to a network endpoint.
    '''
    importer = staticmethod(__import__)

    def __init__(self, func, use_thread=True, use_celery=False,
                 thread_worker=None, *args, **kwargs):
        '''Choose the way how async logging perform.

        :param use_thread: Set to use the asynchronous provided by
            ThreadPoolExecutor. True by default.
        :param use_celery: Set if Celery is available, False by default.
        :param thread_worker: Max workers of ThreadPoolExecutor. If it is None
            or not given, it will default to the number of processors on
            the machine.
        '''
        if use_thread and use_celery:
            raise ValueError('Can not both use thread and celery.')
        elif not use_thread and not use_celery:
            raise ValueError(
                'None of approach are given, set either use_thread or '
                'use_celery to True.')

        if isinstance(func, str):
            func = self.resolve(func)
        if use_thread and not callable(func):
            raise ValueError(
                'func must be a callable function while use_thread is True.')
        elif use_celery and not hasattr(func, 'delay'):
            raise ValueError(
                'Makesure Celery is available while use_celery is True.')

        if thread_worker and not isinstance(thread_worker, int):
            raise ValueError('Integer expected for thread_worker argument.')

        self.func = func
        self.use_thread = use_thread
        self.use_celery = use_celery
        if use_thread:
            self._thread_executor = futures.ThreadPoolExecutor(
                max_workers=thread_worker)
        else:
            self._thread_executor = None
        super(AsyncLogDispatcher, self).__init__(*args, **kwargs)

    def resolve(self, s):
        name = s.split('.')
        used = name.pop(0)
        try:
            found = self.importer(used)
            for frag in name:
                used += '.' + frag
                try:
                    found = getattr(found, frag)
                except AttributeError:
                    self.importer(used)
                    found = getattr(found, frag)
            return found
        except ImportError:
            e, tb = sys.exc_info()[1:]
            v = ValueError('Cannot resolve %r: %s' % (s, e))
            v.__cause__, v.__traceback__ = e, tb
            raise v

    def close(self):
        '''
        Close ThreadPoolExecutor if needed.
        '''
        if self.use_thread and self._thread_executor:
            self._thread_executor.shutdown(wait=True)
        super(AsyncLogDispatcher, self).close()

    def emit(self, record):
        msg = self.format(record)
        if self.use_thread and self._thread_executor:
            self._thread_executor.submit(self.func, msg)
        elif self.use_celery:
            self.func.delay(msg)



from setuptools import setup, find_packages
from os import path
from codecs import open

from asynclog import __version__

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='asynclog',
    version=__version__,
    description='Asynchronous log for python logging.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    author='unpluggedcoder',
    author_email='unpluggedcoder@outlook.com',
    url='https://github.com/unpluggedcoder/asynclog',
    packages=find_packages(),
    package_data={'asynclog': [
        'README.md',
    ]},
    keywords='logging asynchronous',
    install_requires=[],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Logging',
    ], )



import sys
import logging.config
import unittest
from unittest import mock
from functools import wraps

from asynclog import AsyncLogDispatcher

has_celery = False
app = None

try:
    import celery
    from celery import shared_task
    app = celery.Celery('tasks', broker='redis://localhost:6379/0')
    has_celery = True
except ImportError:
    def shared_task(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper


msgs = []


def write_func(msg):
    global msgs
    msgs.append(msg)


@shared_task
def write_task(msg):
    # Write log into network endpoint
    print(msg)


class TestAsyncLog(unittest.TestCase):
    ''' TestCase for AsyncLogDispatcher class
    '''

    def setUp(self):
        global msgs
        msgs = []

    def test_create_asynclog_with_error_params(self):
        with self.assertRaises(ValueError) as ex:
            AsyncLogDispatcher(write_func, use_thread=True, use_celery=True)
        self.assertEqual(ex.exception.args[0],
                         'Can not both use thread and celery.')

        with self.assertRaises(ValueError) as ex:
            AsyncLogDispatcher(write_func, use_thread=False, use_celery=False)
        self.assertEqual(
            ex.exception.args[0],
            'None of approach are given, set either use_thread or '
            'use_celery to True.')

        with self.assertRaises(ValueError) as ex:
            wrong_type = ('wrong', 'type')
            AsyncLogDispatcher(wrong_type, use_thread=True)
        self.assertEqual(
            ex.exception.args[0],
            'func must be a callable function while use_thread is True.')

        with self.assertRaises(ValueError) as ex:
            AsyncLogDispatcher(write_func, use_thread=False, use_celery=True)
        self.assertEqual(
            ex.exception.args[0],
            'Makesure Celery is available while use_celery is True.')

        with self.assertRaises(ValueError) as ex:
            AsyncLogDispatcher(
                write_func,
                use_thread=True,
                use_celery=False,
                thread_worker='2')
        self.assertEqual(ex.exception.args[0],
                         'Integer expected for thread_worker argument.')

    @mock.patch('asynclog.tests.test_handler.write_task')
    def test_can_create_asynclog_with_right_params(self, write_mock):
        instance = AsyncLogDispatcher(
            write_func, use_thread=True, use_celery=False)
        self.assertIsNotNone(instance._thread_executor)
        self.assertIsNotNone(instance.func)

        write_mock.delay = mock.MagicMock()
        instance = AsyncLogDispatcher(
            write_task, use_thread=False, use_celery=True)
        self.assertIsNone(instance._thread_executor)
        self.assertIsNotNone(instance.func)

    def test_can_work_with_thread(self):
        logger = logging.getLogger()
        handler = AsyncLogDispatcher(write_func, use_thread=True)
        handler.setLevel(logging.INFO)
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)
        logger.info('Debug log')
        logger.info('Info log')
        logger.info('Warning log')
        handler.close()
        self.assertEqual(len(msgs), 3)

    @mock.patch('asynclog.tests.test_handler.write_task')
    def test_can_work_with_celery(self, write_mock):
        msg_records = []

        def side_effect(msg):
            msg_records.append(msg)
        write_mock.delay = mock.MagicMock(side_effect=side_effect)
        logger = logging.getLogger()
        handler = AsyncLogDispatcher(
            write_task, use_celery=True, use_thread=False)
        handler.setLevel(logging.INFO)
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)

        logger.info('Debug log')
        logger.info('Info log')
        logger.info('Warning log')
        handler.close()
        if sys.version_info > (3, 6):
            write_mock.delay.assert_called()
        write_mock.delay.assert_has_calls([
            mock.call('Debug log'), mock.call('Info log'),
            mock.call('Warning log')
        ])
        self.assertEqual(len(msg_records), 3)

    def test_can_used_from_dictConfig(self):
        log_cfg = {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'simple': {
                    'format': '%(asctime)s \n %(levelname)s \n %(message)s'
                },
            },
            'handlers': {
                'async_handler': {
                    'level': 'INFO',
                    'formatter': 'simple',
                    'class': 'asynclog.AsyncLogDispatcher',
                    'func': 'asynclog.tests.test_handler.write_func',
                }
            },
            'loggers': {
                'asynclogger': {
                    'handlers': ['async_handler', ],
                    'level': 'DEBUG',
                    'propagate': False,
                },
            }
        }
        logging.config.dictConfig(log_cfg)
        logger = logging.getLogger('asynclogger')
        logger.info('Test1')
        logger.info('Test2')
        logger.info('Test3')
        self.assertEqual(len(msgs), 3)

    def test_error_import_from_dictConfig(self):
        log_cfg = {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'simple': {
                    'format': '%(asctime)s \n %(levelname)s \n %(message)s'
                },
            },
            'handlers': {
                'async_handler': {
                    'level': 'INFO',
                    'formatter': 'simple',
                    'class': 'asynclog.AsyncLogDispatcher',
                    'func': 'asynclog.tests.noone',
                }
            },
            'loggers': {
                'asynclogger': {
                    'handlers': ['async_handler', ],
                    'level': 'DEBUG',
                    'propagate': False,
                },
            }
        }
        with self.assertRaises(ValueError) as ex:
            logging.config.dictConfig(log_cfg)
        self.assertIn(
            'Unable to configure handler \'async_handler\'',
            ex.exception.args[0])


if __name__ == '__main__':
    unittest.main()
